"""Basic functionality for writing datafiles, extended by individual writers

Description:
------------

This module contains an abstract base class that can be extended for writing data files
in Posetta.
"""

# Standard library imports
import codecs
from typing import IO

# Third party imports

# Posetta imports
from posetta import data
from posetta.lib import exceptions


class Writer:
    """An abstract base class that has basic methods for writing a datafile

    This class provides functionality for writing a file. You should inherit from one of
    the specific writers like for instance ChainWriter, LineWriter, SinexWriter etc

    Attributes:
        output_stream: IO[str]    - Stream that output is written to.
        data: data.CoordSet       - The coordinate data to be written.
        writer_name: str          - Name of the writer (module).
        file_path: str            - Name of the datafile that will be written.
        encoding: str             - Encoding of output file.
    """

    def __init__(
        self, output_stream: IO[bytes], cset: data.CoordSet, encoding: str = "utf-8"
    ) -> None:
        """Set up the basic information needed by the writer

        Args:
            output_stream:  Byte stream to write to.
            cset:           Data that will be written.
            encoding:       Encoding used when writing data.
        """
        self.output_stream = codecs.getwriter(encoding)(output_stream)
        self.data = cset
        self.encoding = encoding
        self.writer_name = self.__module__.split(".")[-1]

        try:
            self.file_path = output_stream.name
        except AttributeError:
            self.file_path = "<unknown>"

    def setup_writer(self) -> None:
        """Set up a writer so that it can write data to a file.

        This method may be overwritten if a writer needs to do some preparatory work.
        """
        pass

    def write(self) -> None:
        """Write data

        This is a basic implementation that carries out the whole pipeline of writing
        datafiles.

        Subclasses should typically implement (at least) the `write_data`-method.
        """
        self.setup_writer()
        if self.data.num_obs:
            self.write_data()
        else:
            raise exceptions.WriterError("Input dataset is empty")

    def write_data(self) -> None:
        """Write data to the data file

        Data should be write to `self.file_path` and stored in the dictionary
        `self.data`. A description of the data may be placed in the dictionary
        `self.meta`. If the file is not found, a FileNotFoundError should be raised.
        """
        raise NotImplementedError(f"{self.writer_name} must implement write_data()")

    def __repr__(self) -> str:
        """A simple string representation of the writer
        """
        return f"{self.__class__.__name__}('{self.file_path}')"



"""Posetta, translating between different position formats

Posetta is a command line and GUI utility for translating between different
file formats used for representing points and possibly velocities.

See README.md and __main__.py for more information.
"""

# Standard library imports
from datetime import date as _date
from collections import namedtuple as _namedtuple


# Version of Posetta
#
# This is automatically set using the bumpversion script
__version__ = "0.0.1"


# Start of development
_birthday = _date(2018, 4, 20)


# Maintainers of Posetta
_Author = _namedtuple("_Author", ["name", "email", "start", "end"])
_AUTHORS = [
    _Author("Geir Arne Hjelle", "geir.arne.hjelle@kartverket.no", _birthday, _date.max),
    _Author("Kristian Evers", "kbevers@sdfe.dk", _birthday, _date.max),
]

__author__ = ", ".join(a.name for a in _AUTHORS if a.start < _date.today() < a.end)
__contact__ = ", ".join(a.email for a in _AUTHORS if a.start < _date.today() < a.end)
__url__ = "https://github.com/NordicGeodesy/posetta"


# Copyleft of the library
__copyright__ = f"{_birthday.year} - {_date.today().year} Nordic Geodetic Commision"



"""Coordinate dataset

Description:
------------

The CoordSet is a data structure that is used internally when translating between
different coordinate data formats.

"""

# Standard library imports
from collections import namedtuple
from typing import cast, Dict, List, Optional

# Third party imports
import numpy as np
import pandas as pd

# Posetta imports
from posetta.data import coordset_meta
from posetta.lib import exceptions


# Tables of the CoordSet
_TableSpec = namedtuple("_TableSpec", ("name", "num_cols"))
_TABLES = (
    _TableSpec("epochs", 1),
    _TableSpec("positions", 3),
    _TableSpec("velocities", 3),
    _TableSpec("values", None),
)

# Column specification
ColumnSpec = namedtuple("ColumnSpec", ("table", "idx"))


class CoordSet:
    """Coordinate Dataset

    Attributes:
        num_obs:    Number of rows in dataset.
        meta:       Metainformation about the dataset.
    """

    def __init__(self) -> None:
        """Set up tables for the CoordSet
        """
        self.num_obs = 0
        self.meta = coordset_meta.CoordSetMeta()
        self.columns: Dict[str, ColumnSpec] = dict()
        self.tables = {t.name: Table(t.num_cols) for t in _TABLES}
        for table_name, table in self.tables.items():
            setattr(self, table_name, table)

    def add(
        self,
        table_name: str,
        val: np.ndarray,
        idx: Optional[int] = None,
        column_name: str = "",
    ) -> None:
        """Add one column of values to the CoordSet

        Possible tables are:

            epochs(1), positions(3), velocities(3) and values(None).

        where the number of available columns is listed in the parentheses.

        When adding a column the idx parameter must be between 0 and the above mentioned
        number of colums. If adding a 'value' the index should be set to None, which
        creates a generic table that can hold any number of data columns.
        """
        # Check that number of observations are consistent
        val_num_obs = len(val)
        if not self.columns:
            self.num_obs = val_num_obs
        elif val_num_obs != self.num_obs:
            raise exceptions.CoordSetError(
                f"The number of values of {column_name} are not "
                f"consistent with data already in the CoordSet"
            )

        # Add value to table at specified index
        idx = self.tables[table_name].add_column(column_name, val, idx)

        # Update meta information
        self.columns[column_name] = ColumnSpec(table_name, idx)

    def as_dict(self) -> Dict[str, np.ndarray]:
        """Return columns of coordinate dataset as dictionary
        """
        data = dict()
        for column, table in self.columns.items():
            values = self.tables[table.table].values
            if values is not None:
                data[column] = values[:, table.idx]
        return data

    def as_dataframe(self) -> pd.DataFrame:
        """Return coordinate dataset as Pandas DataFrame
        """
        return pd.DataFrame(self.as_dict())

    def __repr__(self) -> str:
        """A simple string representation of the CoordSet
        """
        tables = [n for n, t in self.tables.items() if t.values is not None]
        return f"{self.__class__.__name__}({', '.join(tables)})"


class Table:
    """Table with columns used in CoordSet
    """

    def __init__(self, num_cols: Optional[int] = None) -> None:
        """Prepare one table
        """
        self._is_expandable = num_cols is None
        self.num_cols = num_cols or 0
        self.values = np.full((0, self.num_cols), np.nan)
        self.col_names = [""] * self.num_cols

    @property
    def has_data(self) -> bool:
        """Has any data been added to the table?
        """
        return self.values.size > 0

    def add_column(self, name: str, val: np.ndarray, idx: Optional[int] = None) -> int:
        """Add one column of values to the Table
        """
        if type(val) is not np.ndarray:
            val = np.asarray(val)

        # Create new array to hold values if necessary
        val_num_obs = len(val)
        if not self.has_data:
            self.values = np.full((val_num_obs, self.num_cols), np.nan)

        # Add value to table at specified index
        if idx is None:
            if self._is_expandable:
                idx = self.values.shape[1] or 0
                self.values = np.concatenate((self.values, val[:, None]), axis=1)
                self.col_names.append(name)
                self.num_cols += 1
            else:
                raise ValueError(f"Table is not expandable, specify 'idx'")
        else:
            self.values[:, idx] = val
            self.col_names[idx] = name

        return idx

    def __repr__(self) -> str:
        """A simple string representation of the Table
        """
        if self.values is None:
            return f"{self.__class__.__name__}()"
        else:
            return f"{self.__class__.__name__}({self.values.shape})"



"""Readers that can read different coordinate file formats

Description:
------------

To add a new reader, simply create a new .py-file which defines a class
inheriting from `posetta.readers._reader.Reader` or one of its subclasses. The
class must be decorated with the `posetta.lib.plugins.register` decorator as
follows:

    from posetta.readers import _reader
    from posetta.lib import plugins

    @plugins.register
    class MyNewFormat(_reader.Reader):
        ...

To use a reader, call it using one of the read-functions defined below:

    from posetta import readers
    data = readers.read_file("file_with_new_format.new", "my_new_format", ...)

or

    from posetta import readers
    with open("file_with_new_format.new", mode="rb") as input_stream:
        data = readers.read_stream(input_stream, "my_new_format", ...)

Note that the stream should be opened as a binary stream to allow the readers
to handle encodings if necessary.


The name used in `read_file` and `read_stream` to call the reader is the name
of the module (file) containing the reader.

"""

# Standard library imports
import pathlib
from typing import Any, IO, List, Optional, Tuple, Union

# Posetta imports
from posetta.readers._reader import Reader
from posetta.lib import exceptions
from posetta.lib import plugins


def names() -> Tuple[str, ...]:
    """List the names of available readers

    Note that this will import all readers.

    Returns:
        Names of the available readers.
    """
    return plugins.list_all(package_name=__name__)


def exists(reader_name: str) -> bool:
    """Check whether the given reader exists

    Args:
        reader_name:  Name of reader.

    Returns:
        True if reader exists, False otherwise.
    """
    return plugins.exists(package_name=__name__, plugin_name=reader_name)


def short_docs(*readers: str) -> List[Tuple[str, str]]:
    """Get one line documentation for readers

    If no readers are specified, documentation for all available readers are returned.

    Args:
        readers:  Names of readers.

    Returns:
        Names and documentation for readers.
    """
    if not readers:
        readers = names()

    return [(r, plugins.doc(__name__, r, long_doc=False)) for r in readers]


def read_stream(
    input_stream: IO[bytes], reader_name: Optional[str] = None, **reader_args: Any
) -> Reader:
    """Read a bytes stream with a given reader

    If the reader is not specified, an attempt to guess at an appropriate
    reader is made. A NoReaderFound error is raised if no such appropriate
    reader is found.

    Args:
        input_stream:  Stream of bytes that should be read.
        reader_name:   Name of reader that should be used.

    Returns:
        Data in stream.
    """
    if reader_name is None:
        reader_name = identify(input_stream)

    reader = plugins.call(
        package_name=__name__,
        plugin_name=reader_name,
        input_stream=input_stream,
        **reader_args,
    )
    reader.read()
    return reader


def read_file(
    file_path: Union[str, pathlib.Path],
    reader_name: Optional[str] = None,
    **reader_args: Any,
) -> Reader:
    """Read a file with a given reader

    If the reader is not specified, an attempt to guess at an appropriate
    reader is made. A NoReaderFound error is raised if no such appropriate
    reader is found.

    Args:
        file_path:    Path to file that should be read.
        reader_name:  Name of reader that should be used.

    Returns:
        Data in file.
    """
    with open(file_path, mode="rb") as input_stream:
        return read_stream(input_stream, reader_name)


def identify(input_stream: IO[bytes]) -> str:
    """Identify a reader that can read a given file

    A NoReaderFound error is raised if no such appropriate reader is found.

    Args:
        file_path:    Path to file that should be identified.

    Returns:
        Name of reader that can read the given file.
    """
    import IPython

    where_am_i = "readers.__init__.identify"
    IPython.embed()

    raise exceptions.NoReaderFound(
        f"Found no reader that can read {input_stream.name}"
    ) from None



"""Writers that can write different coordinate file formats

Description:
------------

To add a new writer, simply create a new .py-file which defines a class
inheriting from `posetta.writers._writer.Writer` or one of its subclasses. The
class must be decorated with the `posetta.lib.plugins.register` decorator as
follows:

    from posetta.writers import _writer
    from posetta.lib import plugins

    @plugins.register
    class MyNewFormat(_writer.Writer):
        ...

To use a writer, call it using one of the write-functions defined below:

    from posetta import writers
    writers.write_file("file_with_new_format.new", "my_new_format", cset, ...)

or

    from posetta import writers
    with open("file_with_new_format.new", mode="wb") as output_stream:
        writers.write_stream(output_stream, "my_new_format", cset, ...)

Note that the stream should be opened in binary mode to allow the writers to
handle encodings if necessary.

The name used in `write_file` and `write_stream` to call the writer is the name
of the module (file) containing the writer.

"""

# Standard library imports
import pathlib
from typing import Any, IO, List, Tuple, Union

# Posetta imports
from posetta.data import CoordSet
from posetta.lib import plugins


def names() -> Tuple[str, ...]:
    """List the names of available writers

    Note that this will import all writers.

    Returns:
        Names of the available writers.
    """
    return plugins.list_all(package_name=__name__)


def exists(writer_name: str) -> bool:
    """Check whether the given writer exists

    Args:
        writer_name:  Name of writer.

    Returns:
        True if writer exists, False otherwise.
    """
    return plugins.exists(package_name=__name__, plugin_name=writer_name)


def short_docs(*readers: str) -> List[Tuple[str, str]]:
    """Get one line documentation for readers

    If no readers are specified, documentation for all available readers are returned.

    Args:
        readers:  Names of readers.

    Returns:
        Names and documentation for readers.
    """
    if not readers:
        readers = names()

    return [(r, plugins.doc(__name__, r, long_doc=False)) for r in readers]


def write_stream(
    output_stream: IO[bytes], writer_name: str, cset: CoordSet, **writer_args: Any
) -> None:
    """Write data to a file with a given writer

    Args:
        output_stream:  Stream of bytes to write to.
        writer_name:    Name of writer that should be used.
        cset:           Data that should be written as a coordinate dataset.
    """
    writer = plugins.call(
        package_name=__name__,
        plugin_name=writer_name,
        output_stream=output_stream,
        cset=cset,
        **writer_args
    )
    writer.write()


def write_file(
    file_path: Union[str, pathlib.Path],
    writer_name: str,
    cset: CoordSet,
    **writer_args: Any
) -> None:
    """Write data to a file with a given writer

    Args:
        file_path:    Path to file that should be written.
        writer_name:  Name of writer that should be used.
        cset:         Data that should be written as a coordinate dataset.
    """
    with open(file_path, mode="wb") as output_stream:
        write_stream(output_stream, writer_name, cset, **writer_args)



"""Gridded data in three columns of numbers, header possible

Description:
------------


Example:
--------

East     North   Value
-25.0000 63.0000 0.0188
-23.0000 63.0000 0.01884
-21.0000 63.0000 0.01798
-19.0000 63.0000 0.01706
-17.0000 63.0000 0.01708
-15.0000 63.0000 0.01795
-13.0000 63.0000 0.01854
-25.0000 65.0000 0.01927
-23.0000 65.0000 0.02014
-21.0000 65.0000 0.02156
-19.0000 65.0000 0.02489
-17.0000 65.0000 0.03102
-15.0000 65.0000 0.01805
-13.0000 65.0000 0.01812


Specification:
--------------

http://www.gdal.org/frmt_xyz.html

"""

# Posetta imports
from posetta import data
from posetta.lib import plugins
from posetta.readers._reader_line import LineReader


@plugins.register
class XyzReader(LineReader):
    """A reader for xyz-files
    """
    _headers = dict(
        x=("positions", 0),
        y=("positions", 1),
        z=("positions", 2),
        east=("positions", 0),
        north=("positions", 1),
        up=("positions", 2),
        lon=("positions", 0),
        lat=("positions", 1),
        height=("positions", 2),
    )

    def setup_reader(self) -> None:
        self.meta["__params__"] = dict(names=True)

    def as_coordset(self) -> data.CoordSet:
        """Return the data as a coordinate dataset

        Returns:
            The data that has been read as a coordinate dataset.
        """
        cset = data.CoordSet()
        for column, value in self.data.items():
            table, idx = self._headers.get(column.lower(), ("values", None))
            cset.add(table, val=value, idx=idx, column_name=column)
        return cset



"""Posetta, The Universal Translator of Geodetic Coordinate File Formats

Posetta can convert input in one format to output with a different
format. Input and output can be either files, or stdin and stdout streams.
"""

# Standard library imports
import pathlib
import tkinter as tk
from tkinter import filedialog
from tkinter import scrolledtext
from tkinter import ttk
from typing import Callable


# Posetta imports
import posetta
from posetta import readers
from posetta import writers


class CallWrapperReportingException(tk.CallWrapper):
    """Overrides the built-in CallWrapper, so that exceptions happening inside tkinter are reported."""

    def __call__(self, *args):
        """Apply first function SUBST to arguments, then FUNC."""
        try:
            if self.subst:
                args = self.subst(*args)
            return self.func(*args)
        except SystemExit:
            raise
        except:  # noqa
            import traceback
            from tkinter import messagebox

            err = traceback.format_exc()
            print(err)
            messagebox.showerror("Oops ...", err)


tk.CallWrapper = CallWrapperReportingException


class PosettaGui(tk.Tk):
    """A simple GUI frontend to the Posetta command line"""

    def __init__(self, translator: Callable, *args, **kwargs):
        # Set up window
        super().__init__(*args, **kwargs)
        super().wm_title("Posetta")

        # Make sure to keep references to icons
        # TODO: How should we handle setting the gui_dir properly
        gui_dir = pathlib.Path(__file__).parent
        self._posetta_icon = tk.PhotoImage(file=gui_dir / "icon.png")
        self.iconphoto(True, self._posetta_icon)

        self._banner = tk.PhotoImage(file=gui_dir / "banner.png")
        self._satellite_icon = tk.PhotoImage(file=gui_dir / "satellite.png")
        self._stone_icon = tk.PhotoImage(file=gui_dir / "stone.png")

        # Callback function for doing translation between files
        self.translator = translator

        # Variables
        self.vars = dict()

        # Layout: Banner image
        banner_line = ttk.Frame(self)
        banner = tk.Label(banner_line, image=self._banner)
        banner.pack(side=tk.TOP, fill=tk.BOTH, expand=False)
        banner_line.pack(side=tk.TOP, fill=tk.X, expand=False)

        # Layout: From file
        from_file = FileAndFormat(
            self,
            text="Input format and file name:",
            fmt_list=readers.names(),
            var_prefix="reader",
            vars=self.vars,
            ask_overwrite=False,
        )
        from_file.pack(side=tk.TOP, fill=tk.X, expand=True)

        # Layout: To file
        to_file = FileAndFormat(
            self,
            text="Output format and file name:",
            fmt_list=writers.names(),
            var_prefix="writer",
            vars=self.vars,
            ask_overwrite=True,
        )
        to_file.pack(side=tk.TOP, fill=tk.X, expand=True)

        # Layout: Controls
        control_line = ttk.Frame(self)
        convert = ttk.Button(
            control_line,
            image=self._satellite_icon,
            text="Convert",
            command=self.translate,
            compound=tk.LEFT,
        )
        convert.pack(side=tk.RIGHT, padx=5, pady=5)
        control_line.pack(side=tk.TOP, fill=tk.X, expand=False)

        # Layout: Status frame
        self.status = Status(self)
        self.status.pack(side=tk.TOP, fill=tk.X, expand=False)

    def translate(self):
        # Gather necessary info
        file_from = self.vars["reader_file"].get()
        if not file_from:
            self.status.write("Please choose input file")
            return

        file_to = self.vars["writer_file"].get()
        if not file_to:
            self.status.write("Please choose output file")
            return

        fmt_from = self.vars["reader_fmt"].get()
        if fmt_from == "detect":
            fmt_from = None

        fmt_to = self.vars["writer_fmt"].get()
        if fmt_to == "detect":
            fmt_to = None

        options = dict(overwrite=True)

        # Call translator
        name_from = pathlib.Path(file_from).name
        name_to = pathlib.Path(file_to).name
        self.status.write(f"Converting {name_from} to {name_to}")
        try:
            self.translator(file_from, file_to, fmt_from, fmt_to, options)
        except Exception as err:
            self.status.write(f"The conversion failed:\n  {err}")
        else:
            self.status.write(f"Done. Output stored in {file_to}\n")


class FileAndFormat(ttk.Frame):

    def __init__(self, master, text, fmt_list, var_prefix, vars, ask_overwrite=False):
        super().__init__(master)
        self.vars = vars
        self.ask_overwrite = ask_overwrite

        # Add autodetection of format
        # fmt_list = ("detect", ) + fmt_list

        # Label
        label_line = ttk.Frame(self)
        tk.Label(label_line, text=text).pack(side=tk.LEFT, padx=5, pady=5)
        label_line.pack(side=tk.TOP, fill=tk.X, expand=False)

        # Format Dropdown
        file_line = ttk.Frame(self)
        fmt_var = self.vars[f"{var_prefix}_fmt"] = tk.StringVar()
        fmt_var.set(fmt_list[0])
        fmt_dropdown = ttk.Combobox(
            file_line, width=12, textvariable=fmt_var, values=fmt_list, state="readonly"
        )
        fmt_dropdown.pack(side=tk.LEFT, padx=20)

        # File Chooser
        self.file_var = self.vars[f"{var_prefix}_file"] = tk.StringVar()
        file_entry = tk.Entry(file_line, width=80, textvariable=self.file_var)
        file_entry.pack(side=tk.LEFT, padx=20)
        file_button = ttk.Button(
            file_line,
            image=self.master._stone_icon,
            text="Choose File",
            command=self._choose_file,
            compound=tk.LEFT,
        )
        file_button.pack(side=tk.LEFT, padx=10)
        file_line.pack(side=tk.BOTTOM, expand=False)

    def _choose_file(self):
        if self.ask_overwrite:
            self.file_var.set(filedialog.asksaveasfilename())
        else:
            self.file_var.set(filedialog.askopenfilename())


class Status(ttk.Frame):

    def __init__(self, master):
        super().__init__(master)

        # Lay out widgets
        self._text = scrolledtext.ScrolledText(self, height=8, wrap=tk.WORD)
        self._text.pack(fill=tk.X, expand=False)

        # Initialize widgets
        startup_text = (
            f"Posetta, v{posetta.__version__}, MIT License {posetta.__copyright__}\n"
            f"See {posetta.__url__} for more information\n"
        )
        self._text.insert(1.0, startup_text)
        self._text.config(state=tk.DISABLED)

    def clear(self):
        self._text.config(state=tk.NORMAL)
        self._text.delete(1.0, tk.END)
        self._text.config(state=tk.DISABLED)

    def write(self, text):
        self._text.config(state=tk.NORMAL)
        self._text.insert(
            tk.END, f"\n{text}"
        )  # New-line in front to avoid blank line at end
        self._text.yview(tk.END)  # Move scrollbar to bottom
        self._text.config(state=tk.DISABLED)



"""Definition of Posetta-specific exceptions

Description:
------------

Custom exceptions used by Posetta for more specific error messages and
handling.
"""


class PosettaException(Exception):
    pass


class CoordSetError(PosettaException):
    pass


class NoReaderFound(PosettaException):
    pass


class ReaderError(PosettaException):
    pass


class WriterError(PosettaException):
    pass


class UnknownPackageError(PosettaException):
    pass


class UnknownPluginError(PosettaException):
    pass



"""Basic functionality for reading datafiles, extended by individual readers

Description:
------------

This module contains an abstract base class that can be extended for reading data files
in Posetta.
"""

# Standard library imports
import codecs
from typing import Any, Dict, IO, List, Union

# Third party imports
import pandas as pd

# Posetta imports
from posetta import data


class Reader:
    """An abstract base class with basic methods for reading a datafile

    This class provides functionality for reading a file. You should
    inherit from one of the specific readers like for instance
    ChainReader, LineReader, SinexReader etc

    Attributes:
        input_stream: IO[str]    - Stream that input is read from.
        reader_name: str         - Name of the reader (module).
        data: Dict[str, Any]     - The coordinate data read from file.
        meta: Dict[str, Any]     - Metainformation read from file.
        file_path: str           - Name of datafile being read.
        encoding: str            - Encoding used when reading.

    """

    def __init__(self, input_stream: IO[bytes], encoding: str = "utf-8") -> None:
        """Set up the basic information needed by the reader

        Args:
            input_stream:  Byte stream that will be read.
            encoding:      Encoding of input stream.
        """
        self.input_stream = codecs.getreader(encoding)(input_stream)
        self.encoding = encoding
        self.reader_name = self.__module__.split(".")[-1]

        try:
            self.file_path = input_stream.name
        except AttributeError:
            self.file_path = "<unknown>"

        # Initialize the data
        self.meta: Dict[str, Any] = dict(
            __reader_name__=self.reader_name,
            __source_path__=self.file_path,
            __source_encoding__=encoding,
        )
        self.data: Dict[str, Any] = dict()

    def setup_reader(self) -> None:
        """Set up a reader so that it can read data from a file.

        This method may be overwritten if a reader needs to do some
        preparatory work.
        """
        pass

    def read(self) -> None:
        """Read data

        This is a basic implementation that carries out the whole
        pipeline of reading datafiles.

        Subclasses should typically implement (at least) the
        `read_data`-method.
        """
        self.setup_reader()
        self.read_data()

    def read_data(self) -> None:
        """Read data from the data file

        Data should be read from `self.input_stream` and stored in the
        dictionary `self.data`. A description of the data may be placed
        in the dictionary `self.meta`.
        """
        raise NotImplementedError(f"{self.reader_name} must implement read_data()")

    def as_dict(self, include_meta: bool = False) -> Dict[str, Any]:
        """Return the data as a dictionary

        Args:
            include_meta:   Include meta-data in the returned dict?

        Returns:
            The data that has been read as a dictionary.
        """
        return dict(self.data, __meta__=self.meta) if include_meta else self.data.copy()

    def as_dataframe(self, index: Union[None, str, List[str]] = None) -> pd.DataFrame:
        """Return the data as a Pandas DataFrame

        This is a basic implementation, assuming the `self.data`
        dictionary has a simple structure. More advanced readers may
        need to reimplement this method.

        Args:
            index:      Name of field to use as index. May also be a list of strings.

        Returns:
            The data that has been read as a DataFrame.

        """
        df = pd.DataFrame.from_dict(self.data)
        if index is not None:
            df.set_index(index, drop=True, inplace=True)

        return df

    def as_coordset(self) -> data.CoordSet:
        """Return the data as a coordinate dataset

        Returns:
            The data that has been read as a coordinate dataset.
        """
        raise NotImplementedError(f"{self.reader_name} must implement as_coordset()")

    def __repr__(self) -> str:
        """A simple string representation of the reader
        """
        return f"{self.__class__.__name__}('{self.file_path}')"



#!/usr/bin/env python3
"""Posetta, The Universal Translator of Geodetic Coordinate File Formats

Posetta can convert input in one format to output with a different
format. Input and output can be either files, or stdin and stdout streams.

\b
Examples:
---------

Convert file_1.txt in XYZ-format to file_2.gri in GRI-format:

  $ posetta -f file_1.txt -F xyz -t file_2.gri -T gri

\b
Input Formats:
--------------

\b
{readers}

\b
Output formats:
---------------

\b
{writers}

\b
About Posetta:
--------------

Posetta, v{version}, MIT License {copyright}

Currently maintained by:

\b
{maintainers}

Contributions are welcome at {url}.

"""

# Standard library imports
import pathlib
import sys
from typing import Any, Dict, Optional, Union

# Third party imports
import click

# Posetta imports
import posetta
from posetta import readers
from posetta import writers


def help_str() -> str:
    """Add information to the module doc-string for a complete help message
    """
    maintainers = [
        f"  + {name} <{email}>"
        for name, email in zip(
            posetta.__author__.split(", "), posetta.__contact__.split(", ")
        )
    ]
    return __doc__.format(
        readers="\n".join(f"  + {name} - {doc}" for name, doc in readers.short_docs()),
        writers="\n".join(f"  + {name} - {doc}" for name, doc in writers.short_docs()),
        maintainers="\n".join(maintainers),
        url=posetta.__url__,
        copyright=posetta.__copyright__,
        version=posetta.__version__,
    )


#
# Starting Point of Command Line Tool
#
@click.command(help=help_str())
@click.option("-f", "--file_from", help="Path of input file. Default is stdin.")
@click.option("-t", "--file_to", help="Path of output file. Default is stdout.")
@click.option(
    "-F", "--fmt_from", help="Input file format (supported formats listed above)."
)
@click.option(
    "-T", "--fmt_to", help="Output file format (supported formats listed above)."
)
@click.option(
    "-O", "--overwrite", is_flag=True, help="Overwrite if output file already exists."
)
@click.option("-V", "--verbose", is_flag=True, help="Say what is happening.")
@click.version_option(None, "-v", "--version")
@click.help_option("-h", "--help")
def cli(
    file_from: str, file_to: str, fmt_from: str, fmt_to: str, **options: Dict[str, Any]
) -> None:
    """The command line interface for Posetta

    Implemented using click: http://click.pocoo.org
    """
    translate(file_from, file_to, fmt_from, fmt_to, options)


#
# Starting Point for the Graphical (GUI) Tool
#
def gui() -> None:
    from posetta.gui import PosettaGui  # Local import to avoid slowing down command line tool

    root = PosettaGui(translate)
    root.mainloop()


#
# Both CLI and GUI end up here :)
#
def translate(
    file_from: Union[str, pathlib.Path],
    file_to: Union[str, pathlib.Path],
    fmt_from: str,
    fmt_to: str,
    options: Optional[Dict[str, Any]] = None,
) -> None:
    """The main workflow of Posetta

    Read file_from with format fmt_from, translate it to format fmt_to and write
    it to file_to.

    Args:
        file_from:  Name of input file.
        file_to:    Name of output file.
        fmt_from:   Format of input file.
        fmt_to:     Format of output file.
        options:    Additional options.
    """
    options = options or dict()
    verbose = _verbose_on if options.get("verbose") else _verbose_off
    path_from = None if file_from is None else pathlib.Path(file_from)
    path_to = None if file_to is None else pathlib.Path(file_to)

    # Check input parameters
    if path_from is not None and not path_from.exists():
        raise click.BadParameter(f"Input file '{path_from}' does not exist")

    if path_to is not None and path_to.exists() and not options.get("overwrite"):
        raise click.BadParameter(
            f"Output file '{path_to}' already exists. Use --overwrite to overwrite it."
        )

    if fmt_from is not None and not readers.exists(fmt_from):
        fmts = ", ".join(readers.names())
        raise click.BadParameter(
            f"Input format '{fmt_from}' is not supported. Use one of {fmts}"
        )

    if fmt_to is None or not writers.exists(fmt_to):
        fmts = ", ".join(writers.names())
        raise click.BadParameter(
            f"Output format '{fmt_to}' is not supported. Use one of {fmts}"
        )

    # Read input as CoordSet
    if path_from is None:
        verbose("Reading from standard input")
        cset = readers.read_stream(sys.stdin.buffer, fmt_from).as_coordset()
    else:
        verbose(f"Reading from '{path_from}'")
        cset = readers.read_file(path_from, fmt_from).as_coordset()

    # Write CoordSet to new format
    if path_to is None:
        verbose("Writing to standard output")
        writers.write_stream(sys.stdout.buffer, fmt_to, cset)
    else:
        verbose(f"Writing to '{path_to}'")
        writers.write_file(path_to, fmt_to, cset)


#
# Convenience functions
#
def _verbose_on(text: str) -> None:
    """Print text to stdout

    Args:
        text:  Text that should be printed.
    """
    print(text)


def _verbose_off(text: str) -> None:
    """Do nothing

    Args:
        text:  Text that will not be printed.
    """
    pass


if __name__ == "__main__":
    cli()



"""Set up a simple plug-in architecture for Posetta

Description:
------------

In order to be able to add readers and writers etc without needing to hardcode
names, but rather pick them from the command line or similar, we use a simple
plug-in architecture. The plug-in mechanism is based on the different plug-ins
registering themselves using the :func:`register` decorator::

    from posetta.lib import plugins

    @plugins.register
    def simple_reader(file_path, num_points):
        ...

Plug-ins are registered based on the name of the module (file) they are defined
in, as well as the package (directory) which contains them. Typically all
plug-ins of a given type are collected in a package, e.g. readers, writers,
etc. To list all plug-ins in a package use :func:`list_all`::

    > from posetta.lib import plugins
    > plugins.list_all('posetta.readers')
    ['reader_one', 'reader_three', 'reader_two']

Note that the plug-ins are sorted alphabetically.

A plug-in may be called using :func:`call`::

    > from posetta.lib import plugins
    > plugins.call('posetta.readers', plugin_name='reader_one',
    ...            arg_to_plugin='hello')
    <result from reader_one>

Arguments to the plug-ins should be passed as named arguments to :func:`call`.
"""

# Standard library imports
from collections import namedtuple
import importlib
import pathlib
import re
import sys
from typing import Any, Callable, Dict, Tuple

# Posetta imports
from posetta.lib import exceptions


# Simple structure containing information about a plug-in
Plugin = namedtuple("Plugin", ["name", "function", "doc"])
Plugin.__doc__ = """Information about a plug-in

    Args:
        name: str        - Name of the plug-in.
        function:        - The plug-in.
        doc: str         - Doc string of the plug-in (module)
    """


# The _PLUGINS-dict is populated by the :func:`register` decorator in each
# module.
_PLUGINS: Dict[str, Dict[str, Plugin]] = dict()


#
# REGISTER PLUG-INS
#
def register(func: Callable) -> Callable:
    """Register a plug-in

    Plug-ins are registered based on the name of the module (file) they are
    defined in, as well as the package (directory) which contains
    them. Typically all plug-ins of a given type are collected in a package,
    e.g. readers, writers, etc.

    Args:
        func:  The function that is being registered.

    Returns:
        The function that is being registered (untouched).
    """
    # Get information from the function being registered
    package_name, _, plugin_name = func.__module__.rpartition(".")
    doc = sys.modules[func.__module__].__doc__
    plugin = Plugin(plugin_name, func, doc or "")

    # Store Plugin-object in _PLUGINS dictionary
    _PLUGINS.setdefault(package_name, dict())[plugin_name] = plugin

    return func


#
# CALL PLUG-INS
#
def call(package_name: str, plugin_name: str, **plugin_args: Any) -> Any:
    """Call a plug-in

    If the plug-in is not part of the package an UnknownPluginError is raised.

    Args:
        package_name:  Name of package containing plug-ins.
        plugin_name:   Name of the plug-in (module).
        plugin_args:   Named arguments passed on to the plug-in.

    Returns:
        Return value of the plug-in.
    """
    # Get Plugin-object
    plugin = load(package_name, plugin_name)

    # Call plug-in
    return plugin.function(**plugin_args)


#
# GET DOCUMENTATION FOR PLUG-INS
#
def doc(
    package_name: str,
    plugin_name: str,
    long_doc: bool = True,
    include_details: bool = False,
) -> str:
    """Document one plug-in

    Documentation is taken from the module doc-string. If the plug-in is not part of the
    package an UnknownPluginError is raised.

    Args:
        package_name:     Name of package containing plug-ins.
        plugin_name:      Name of the plug-in (module).
        long_doc:         Use long doc-string or short one-line string.
        include_details:  Include development details like parameters and return values?

    Returns:
        Documentation of the plug-in.
    """
    # Get Plugin-object and pick out doc-string
    doc = load(package_name, plugin_name).doc

    if long_doc:
        # Strip short description and indentation
        lines = [d.strip() for d in "\n\n".join(doc.split("\n\n")[1:]).split("\n")]

        # Stop before Args:, Returns: etc if details should not be included
        idx_args = len(lines)
        if not include_details:
            re_args = re.compile("(Args:|Returns:|Details:|Attributes:)$")
            try:
                idx_args = [re_args.match(l) is not None for l in lines].index(True)
            except ValueError:
                pass
        return "\n".join(lines[:idx_args]).strip()

    else:
        # Return short description
        return doc.split("\n\n")[0].replace("\n", " ").strip()


#
# LIST AVAILABLE PLUG-INS
#
def list_all(package_name: str) -> Tuple[str, ...]:
    """List all plug-ins in a package

    Lists all available plug-ins in the package.  Do note, however, that this
    will import all python files in the package.

    Args:
        package_name:  Name of package containing plug-ins.

    Returns:
        Sorted list of strings with names of plug-ins.
    """
    # Import all plug-ins
    _import_all(package_name)

    # Figure out names of plug-ins
    plugin_names = _PLUGINS.get(package_name, dict()).keys()

    return tuple(sorted(plugin_names))


def exists(package_name: str, plugin_name: str) -> bool:
    """Check whether or not a plug-in exists in a package

    Tries to import the given plug-in.

    Args:
        package_name:  Name of package containing plug-ins.
        plugin_name:   Name of the plug-in (module).

    Returns:
        True if plug-in exists, False otherwise.
    """
    if plugin_name not in _PLUGINS.get(package_name, dict()):
        try:
            _import_one(package_name, plugin_name)
        except exceptions.UnknownPluginError:
            return False

    return plugin_name in _PLUGINS.get(package_name, dict())


#
# LOAD PLUG-INS
#
def load(package_name: str, plugin_name: str) -> Plugin:
    """Load one plug-in from a package

    Tries to load the plugin with the given name.

    Args:
        package_name:  Name of package containing plug-ins.
        plugin_name:   Name of the plug-in (module).

    Returns:
        Name of plug-in.
    """
    if plugin_name not in _PLUGINS.get(package_name, dict()):
        _import_one(package_name, plugin_name)

    try:
        return _PLUGINS[package_name][plugin_name]
    except KeyError:
        msg = "Module '{}.{}' does not contain a plugin".format(
            package_name, plugin_name
        )
        raise exceptions.UnknownPluginError(msg) from None


def _import_one(package_name: str, plugin_name: str) -> None:
    """Import a plugin from a package

    This is essentially just a regular python import. As the module is
    imported, the _PLUGINS-dict will be populated by @register decorated
    function in the file.

    Args:
        package_name:  Name of package containing plug-ins.
        plugin_name:   Name of the plug-in (module).
    """
    try:
        importlib.import_module(package_name + "." + plugin_name)
    except ImportError:
        msg = "Plugin '{}' not found in '{}'".format(plugin_name, package_name)
        raise exceptions.UnknownPluginError(msg) from None


def _import_all(package_name: str) -> None:
    """Import all .py-files in the given package directory

    As each file is imported, the _PLUGINS-dict will be populated by @register
    decorated functions in each file.

    Files with names starting with an underscore are not imported.

    Args:
        package_name:  Name of package containing plug-ins.
    """
    # Figure out the directory of the package by importing it. The package must contain
    # __init__.py for this to work, as otherwise __file__ is not set.
    try:
        package = importlib.import_module(package_name)
    except ImportError:
        msg = f"Plug-in package '{package_name}' not found"
        raise exceptions.UnknownPackageError(msg) from None

    try:
        directory = pathlib.Path(package.__file__).parent
    except AttributeError:
        msg = f"Plug-in package '{package_name}' must include '__init__.py'"
        raise exceptions.UnknownPluginError(msg) from None

    # Import all .py files in the given directory
    for file_path in directory.glob("*.py"):
        plugin_name = file_path.stem
        if not plugin_name.startswith("_"):
            _import_one(package_name, plugin_name)



"""Coordinate dataset meta-information

Description:
------------

"""

# Standard library imports
from typing import Dict, Tuple

# Third party imports

# Posetta imports


class CoordSetMeta:
    """Coordinate Dataset Metainformation


    Attributes:
        fields:   Fields in CoordSet
    """

    def __init__(self) -> None:
        """Set up fields for the CoordSetMeta
        """
        pass

    def __repr__(self) -> str:
        """A simple string representation of the reader
        """
        return f"{self.__class__.__name__}()"



"""XML style format

Description:
------------


Example:
--------



Specification:
--------------

http://www.opengeospatial.org/standards/kml/

"""

from fastkml import kml
import numpy as np
import pandas as pd

# Posetta imports
from posetta import data
from posetta.lib import plugins
from posetta.readers._reader import Reader


@plugins.register
class KmlReader(Reader):
    """A reader for KML-files
    """
    def setup_reader(self) -> None:
        """Set up a reader so that it can read data from a file."""
        self.kml = kml.KML()

    def read_data(self) -> None:
        """Read data from the data file

        Data should be read from `self.input_stream` and stored in the
        dictionary `self.data`. A description of the data may be placed
        in the dictionary `self.meta`.
        """
        try:
            bytes(self.input_stream)
        except TypeError:
            pass  # Figure out how to fix this
        kml_string = self.input_stream.read()
        self.kml.from_string(kml_string)
        for point, description in self._points(self.kml.features()):
            self.data.setdefault("__points__", list()).append(np.array(point))
            for k, v in description.items():
                self.data.setdefault(k, list()).append(v)

    def _points(self, features):
        """Find points and description recursively"""
        for feature in features:
            if hasattr(feature, "geometry"):
                points = feature.geometry
                description = pd.read_html(feature.description)[0]
                description = description.set_index(description[0]).drop(columns=0)
                description = description.to_dict()[1]
                yield points, description
            else:
                yield from self._points(feature.features())

    def as_coordset(self) -> data.CoordSet:
        """Return the data as a coordinate dataset

        Returns:
            The data that has been read as a coordinate dataset.
        """
        cset = data.CoordSet()
        points = np.array(self.data.pop("__points__"))
        num_points = len(points)
        for idx, value in enumerate(points.T):
            cset.add("positions", val=value, idx=idx, column_name="xyz"[idx])

        for column, value in self.data.items():
            if len(value) != num_points:
                continue
            cset.add("values", val=value, column_name=column)

        return cset



"""GUI frontend for Posetta

The GUI frontend gives an alternative for those that are not comfortable using
the command line.

"""

from posetta.gui.posetta_gui import PosettaGui  # type: ignore  # noqa



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['posetta',
 'posetta.data',
 'posetta.gui',
 'posetta.lib',
 'posetta.readers',
 'posetta.writers']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'fastkml>=0.11.0,<0.12.0',
 'lxml>=4.2,<5.0',
 'pandas>=0.23.4,<0.24.0']

setup_kwargs = {
    'name': 'posetta',
    'version': '0.0.1',
    'description': 'The universal translator of geodetic coordinate file formats',
    'long_description': None,
    'author': 'Geir Arne Hjelle',
    'author_email': 'geir.arne.hjelle@kartverket.no',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



"""Simple format readable by PROJ

Description:
------------

PROJ reads files on a line by line basis. Each line contains a coordinate
and possibly a comment. A coordinate is made up of two to four numbers,
usually ordered as x, y, z, t, but this isn't a requirement. Anything
after a # is considered a comment by PROJ. We use that fact to add
extraneous data such as velocities and other generic values.

"""

# Posetta imports
from posetta.lib import exceptions
from posetta.lib import plugins
from posetta.writers._writer import Writer


@plugins.register
class ProjWriter(Writer):
    """A writer of files that PROJ will understand
    """

    def write_data(self) -> None:
        """Write data to a text file in 'x y z t # comment' format.
        """

        # TODO: We can't do tests based on column names unless we standardize them
        # # check that dataset has the required columns
        # required_columns = {"easting": False, "northing": False}
        # for column in self.data.columns:
        #     # print(column.table, column.idx)
        #     if column in ("easting", "northing"):
        #         required_columns[column] = True

        # if not (all(value is True for value in required_columns.values())):
        #     raise exceptions.WriterError(
        #         "Input dataset should at least contain eating and northing columns"
        #     )

        for idx in range(self.data.num_obs):
            # assume that position data is ordered as easting, northing, elevation
            # (otherwise a PROJ axisswap operation can fix that).
            for col_idx in range(self.data.tables["positions"].num_cols):
                self.output_stream.write(
                    "{val}\t".format(
                        val=self.data.tables["positions"].values[idx][col_idx]
                    )
                )

            # write epochs if they are present
            if self.data.tables["epochs"].has_data:
                self.output_stream.write(
                    "{val}\t".format(val=self.data.tables["epochs"].values[idx][0])
                )

            # everything from here is considered a comment by PROJ
            self.output_stream.write("#  ")

            # write velocities if they are present
            if self.data.tables["velocities"].has_data:
                for col_idx in range(self.data.tables["velocities"].num_cols):
                    self.output_stream.write(
                        "{val}\t".format(
                            val=self.data.tables["velocities"].values[idx][col_idx]
                        )
                    )

            # write additional values if they are present
            if self.data.tables["values"].has_data:
                for col_idx in range(self.data.tables["values"].num_cols):
                    self.output_stream.write(
                        "{val}\t".format(
                            val=self.data.tables["values"].values[idx][col_idx]
                        )
                    )

            self.output_stream.write("\n")



"""Loosely formatted coordinate files from KMS

Description:
------------

Read files from old KMS software. The files are generally not structured in a
manner that is particularly readable and the format is very loosely defined.
Hence it is only possible to read a subset of the possible data in a KMS file.

Example files:
--------------

    kms\koordinater
    kms\koter
"""
# Standard library imports
import re
from typing import IO

# Posetta imports
from posetta import data
from posetta.lib import plugins
from posetta.readers._reader_line import Reader


@plugins.register
class KmsReader(Reader):
    """A reader for KMS files
    """

    def __init__(self, input_stream: IO[bytes], encoding: str = "latin-1") -> None:
        """Set up the basic information needed by the reader

        Kms files will usually, if not always, be generated by software
        created before UTF-8 was thought of... So do latin-1 as default
        encoding.

        Args:
            input_stream:  Byte stream that will be read.
        """
        super().__init__(input_stream, encoding)

        fields = [
            "easting",
            "northing",
            "elevation",
            "epoch",
            "minilabel",
            "comment",
            "station",
        ]
        self.data = {f: [] for f in fields}

    def setup_reader(self) -> None:
        self.meta["__params__"] = dict(names=True)

    def read_data(self) -> None:
        """
        Read KMS coordinate file.
        """
        text = self.input_stream.read()
        text = re.sub(r"\*.*?;", "", text, flags=re.DOTALL)  # Remove comment lines
        text = re.sub(r"\s+\n", "\n", text, flags=re.MULTILINE)  # Remove empty lines
        lines = text.split("\n")

        for line in lines:
            if line[0:17].strip().startswith("#"):
                kms_label = line.strip()
            elif line[0:17].strip().startswith("-1z"):
                kms_label = ""
            else:
                northing = None
                easting = None
                elevation = None

                if line and line[0] != " ":
                    continue

                station_id = line[0:18].replace(" ", "")
                # station_descr = line[18:21].replace(" ", "")
                cols = re.split("  +", line[21:].strip())
                if len(cols) == 1 and cols[0] == "":
                    continue

                elements = []
                for i, element in enumerate(cols):
                    value = None
                    unit = None
                    if element.strip().endswith((" m", " dg", " rad")):
                        value = "".join(element.strip().split(" ")[:-1])
                        unit = element.strip().split(" ")[-1]

                    if element.strip().endswith(("\dm", "\ddg", "\drad")):
                        value = re.findall(r"[-+]?\d*\.\d+|\d+", element.strip())[0]
                        unit = re.findall(r"\w$", element.strip())[0]

                    if element.strip().endswith((" sx", " nt")):
                        value = element.strip()
                        unit = element.strip().split(" ")[-1]

                    if value is not None:
                        elements.append(float(value))

                if len(elements) == 3:
                    easting = elements[1]
                    northing = elements[0]
                    elevation = elements[2]

                if len(elements) == 2:
                    easting = elements[1]
                    northing = elements[0]
                    elevation = 0

                if len(elements) == 1:
                    easting = 0
                    northing = 0
                    elevation = elements[0]

                if len(elements) > 0:
                    self.data["comment"].append(" ".join(cols[len(elements):]))
                    self.data["minilabel"].append(kms_label)
                    self.data["station"].append(station_id)
                    if easting is not None:
                        self.data["easting"].append(easting)
                    if northing is not None:
                        self.data["northing"].append(northing)
                    if elevation is not None:
                        self.data["elevation"].append(elevation)

    def as_coordset(self) -> data.CoordSet:
        """Return the data as a coordinate dataset

        Returns:
            The data that has been read as a coordinate dataset.
        """
        cset = data.CoordSet()

        # add coordinates
        if all(v is not None for v in self.data["easting"]):
            cset.add("positions", self.data["easting"], 0, "easting")
        if all(v is not None for v in self.data["northing"]):
            cset.add("positions", self.data["northing"], 1, "northing")
        if all(v is not None for v in self.data["elevation"]):
            cset.add("positions", self.data["elevation"], 2, "elevation")

        # add whatever metadata is available
        if all(v is not None for v in self.data["station"]):
            cset.add("values", self.data["station"], None, "station")
        if all(v is not None for v in self.data["minilabel"]):
            cset.add("values", self.data["minilabel"], None, "minilabel")
        if all(v is not None for v in self.data["comment"]):
            cset.add("values", self.data["comment"], None, "comment")

        return cset



"""Basic functionality for reading datafiles line by line using Numpy

Description:
------------

Read file formats that contain data in nicely formatted columns in text
files.
"""

# Standard library imports
import pathlib
from typing import IO

# Third party imports
import numpy as np

# Posetta imports
from posetta.lib import exceptions
from posetta.readers._reader import Reader


class LineReader(Reader):
    """An abstract base class with basic methods for reading a datafile

    This class provides functionality for using numpy to read a file line by
    line. You should inherit from this one, and at least specify the necessary
    parameters in `setup_reader`.
    """

    def __init__(self, input_stream: IO[bytes], encoding: str = "utf-8") -> None:
        """Set up the basic information needed by the Reader

        Add a self._array property for the raw numpy array data.

        Args:
            input_stream:    Byte stream that will be read.
            encoding:        Encoding of input stream.
        """
        super().__init__(input_stream, encoding)
        self._array = None

    def setup_reader(self) -> None:
        """Set up information needed for the reader

        This method must create a dictionary at `self.meta["__params__"]`
        containing all parameters needed by np.genfromtxt to do the actual
        reading.
        """
        raise NotImplementedError(f"{self.reader_name} must implement setup_reader()")

    def read_data(self) -> None:
        """Read data from the data file

        Uses the np.genfromtxt-function to read the file. Any necessary
        parameters should be returned by `setup_reader`. See
        `self.structure_data` if the self.data-dictionary needs to be
        structured in a particular way.
        """
        if "__params__" not in self.meta:
            raise exceptions.ReaderError(
                f"{self.__class__.__name__} is not properly set up."
            )

        self._array = np.genfromtxt(self.input_stream, **self.meta["__params__"])
        self.structure_data()

    def structure_data(self) -> None:
        """Structure raw array data into the self.data dictionary

        This simple implementation creates a dictionary with one item per
        column in the array. Override this method for more complex use cases.
        """
        if self._array is None:
            raise exceptions.ReaderError(
                f"No data found in {type(self)}. Have you called read_data() yet?"
            )

        for name in self._array.dtype.names:
            self.data[name] = self._array[name]
